@@ -1,16 +1,16 @@
 /*
   Copyright 2015-2016 Oliver Heimlich
-  
+
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.
-  
+
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
-  
+
   You should have received a copy of the GNU General Public License
   along with this program; if not, see <http://www.gnu.org/licenses/>.
 */
@@ -40,16 +40,24 @@ void evaluate (
 {
   mpfr_t mp;
   mpfr_init2 (mp, BINARY64_PRECISION);
-  
+  mpfr_exp_t old_emin = mpfr_get_emin ();
+  mpfr_set_emin (BINARY64_EMIN);
+
   const octave_idx_type n = arg1.numel ();
   for (octave_idx_type i = 0; i < n; i ++)
     {
       mpfr_set_d (mp, arg1.elem (i), MPFR_RNDZ);
-      (*f) (mp, mp, rnd);
+      int rnd_error = (*f) (mp, mp, rnd);
+      if (rnd == MPFR_RNDN)
+        {
+          // Prevent double-rounding errors
+          mpfr_subnormalize (mp, rnd_error, rnd);
+        }
       arg1.elem (i) = mpfr_get_d (mp, rnd);
     }
-  
+
   mpfr_clear (mp);
+  mpfr_set_emin (old_emin);
 }
 
 // Evaluate a binary MPFR function on two binary64 matrices
@@ -62,20 +70,22 @@ void evaluate (
   mpfr_t mp1, mp2;
   mpfr_init2 (mp1, BINARY64_PRECISION);
   mpfr_init2 (mp2, BINARY64_PRECISION);
-  
+  mpfr_exp_t old_emin = mpfr_get_emin ();
+  mpfr_set_emin (BINARY64_EMIN);
+
   // arg1 shall contain the result and must be resized
   if (arg1.rows () == 1 && arg2.rows () != 1)
-    arg1 = arg1.index (idx_vector (ColumnVector (arg2.rows (), 1.0)), 
+    arg1 = arg1.index (idx_vector (ColumnVector (arg2.rows (), 1.0)),
                        idx_vector::colon);
   if (arg1.columns () == 1 && arg2.columns () != 1)
     arg1 = arg1.index (idx_vector::colon,
                        idx_vector (RowVector (arg2.columns (), 1.0)));
-  
+
   const octave_idx_type n = arg1.rows ();
   const octave_idx_type m = arg1.columns ();
   const bool broadcast_r = n != 1 && arg2.rows () == 1;
   const bool broadcast_c = m != 1 && arg2.columns () == 1;
-  
+
   for (octave_idx_type i = 0; i < n; i ++)
     for (octave_idx_type j = 0; j < m; j ++)
       {
@@ -87,12 +97,18 @@ void evaluate (
                       : ((broadcast_c) ? arg2.elem (i, 0)
                                        : arg2.elem (i, j))
                     , MPFR_RNDZ);
-        (*f) (mp1, mp1, mp2, rnd);
+        int rnd_error = (*f) (mp1, mp1, mp2, rnd);
+        if (rnd == MPFR_RNDN)
+          {
+            // Prevent double-rounding errors
+            mpfr_subnormalize (mp1, rnd_error, rnd);
+          }
         arg1.elem (i, j) = mpfr_get_d (mp1, rnd);
       }
-  
+
   mpfr_clear (mp1);
   mpfr_clear (mp2);
+  mpfr_set_emin (old_emin);
 }
 
 // Evaluate a ternary MPFR function on three binary64 matrices
@@ -107,11 +123,13 @@ void evaluate (
   mpfr_init2 (mp1, BINARY64_PRECISION);
   mpfr_init2 (mp2, BINARY64_PRECISION);
   mpfr_init2 (mp3, BINARY64_PRECISION);
-  
+  mpfr_exp_t old_emin = mpfr_get_emin ();
+  mpfr_set_emin (BINARY64_EMIN);
+
   bool scalar1 = arg1.numel () == 1;
   bool scalar2 = arg2.numel () == 1;
   bool scalar3 = arg3.numel () == 1;
-  
+
   if (scalar1)
     {
       // arg1 shall contain the result and must possibly be resized
@@ -140,13 +158,19 @@ void evaluate (
                   (scalar3) ? arg3.elem (0) // broadcast
                             : arg3.elem (i),
                   MPFR_RNDZ);
-      (*f) (mp1, mp1, mp2, mp3, rnd);
+      int rnd_error = (*f) (mp1, mp1, mp2, mp3, rnd);
+      if (rnd == MPFR_RNDN)
+        {
+          // Prevent double-rounding errors
+          mpfr_subnormalize (mp1, rnd_error, rnd);
+        }
       arg1.elem (i) = mpfr_get_d (mp1, rnd);
     }
-  
+
   mpfr_clear (mp1);
   mpfr_clear (mp2);
   mpfr_clear (mp3);
+  mpfr_set_emin (old_emin);
 }
 
 // Evaluate nthroot
@@ -157,7 +181,7 @@ void nthroot (
 {
   mpfr_t mp;
   mpfr_init2 (mp, BINARY64_PRECISION);
-  
+
   const octave_idx_type n = arg1.numel ();
   for (octave_idx_type i = 0; i < n; i ++)
     {
@@ -165,33 +189,62 @@ void nthroot (
       mpfr_root (mp, mp, arg2, rnd);
       arg1.elem (i) = mpfr_get_d (mp, rnd);
     }
-  
+
   mpfr_clear (mp);
 }
 
 // Evaluate factorial
 void factorial (
-  Matrix &result,          // Result
-  const uint64NDArray arg, // Operand
+  Matrix &arg1, // Operand 1 and result
   const mpfr_rnd_t rnd)
 {
   mpfr_t mp;
   mpfr_init2 (mp, BINARY64_PRECISION);
 
-  const octave_idx_type n = arg.numel ();
+  const octave_idx_type n = arg1.numel ();
   for (octave_idx_type i = 0; i < n; i ++)
     {
-      // Compilation on 32-bit systems produces ambiguity errors if the
-      // type of arg.elem(i) is not defined explicitly
-      const uint64_t current_arg = arg.elem (i);
-      mpfr_fac_ui (mp, current_arg, rnd);
-      result.elem (i) = mpfr_get_d (mp, rnd);
+      if (std::isnan (arg1.elem (i)))
+        continue;
+
+      if (arg1.elem (i) < 2.0)
+        {
+          arg1.elem (i) = 1.0;
+          continue;
+        }
+
+      if (arg1.elem (i) >= 171.0)
+        {
+          // Computation can become hard for large numbers,
+          // thus we can short-circuit here.
+          switch (rnd)
+            {
+              case MPFR_RNDZ:
+              case MPFR_RNDD:
+                arg1.elem (i) = std::numeric_limits <double>::max ();
+                continue;
+              case MPFR_RNDA:
+              case MPFR_RNDU:
+              case MPFR_RNDN:
+                arg1.elem (i) = +INFINITY;
+                continue;
+              default:
+                break;
+            }
+        }
+
+      // The factorial function is defined as the product of all positive
+      // integers less than or equal to n.
+      const double current_arg = floor (arg1.elem (i));
+
+      mpfr_fac_ui (mp, static_cast <unsigned long int> (current_arg), rnd);
+      arg1.elem (i) = mpfr_get_d (mp, rnd);
     }
 
   mpfr_clear (mp);
 }
 
-DEFUN_DLD (mpfr_function_d, args, nargout, 
+DEFUN_DLD (mpfr_function_d, args, nargout,
   "-*- texinfo -*-\n"
   "@documentencoding UTF-8\n"
   "@defun mpfr_function_d ('acos', @var{R}, @var{X})\n"
@@ -237,8 +290,6 @@ DEFUN_DLD (mpfr_function_d, args, nargout,
   "@defunx mpfr_function_d ('sech', @var{R}, @var{X})\n"
   "@defunx mpfr_function_d ('sin', @var{R}, @var{X})\n"
   "@defunx mpfr_function_d ('sinh', @var{R}, @var{X})\n"
-  "@defunx mpfr_function_d ('sinh', @var{R}, @var{X})\n"
-  "@defunx mpfr_function_d ('sinh', @var{R}, @var{X})\n"
   "@defunx mpfr_function_d ('sqr', @var{R}, @var{X})\n"
   "@defunx mpfr_function_d ('tan', @var{R}, @var{X})\n"
   "@defunx mpfr_function_d ('tanh', @var{R}, @var{X})\n"
@@ -267,7 +318,7 @@ DEFUN_DLD (mpfr_function_d, args, nargout,
   "  @result{} 1\n"
   "@end group\n"
   "@end example\n"
-  "@seealso{fesetround}\n"
+  "@seealso{crlibm_function}\n"
   "@end defun"
   )
 {
@@ -278,7 +329,7 @@ DEFUN_DLD (mpfr_function_d, args, nargout,
       print_usage ();
       return octave_value_list ();
     }
-  
+
   // Read parameters
   const std::string function = args (0).string_value ();
   const mpfr_rnd_t  rnd      = parse_rounding_mode (args (1).scalar_value ());
@@ -305,7 +356,7 @@ DEFUN_DLD (mpfr_function_d, args, nargout,
     }
   if (error_state)
     return octave_value_list ();
-  
+
   // Choose the function to evaluate
   switch (nargin - 2)
     {
@@ -349,12 +400,7 @@ DEFUN_DLD (mpfr_function_d, args, nargout,
         else if (function == "expm1")
           evaluate (arg1, rnd, &mpfr_expm1);
         else if (function == "factorial")
-          {
-            const uint64NDArray argInt = args (2).uint64_array_value ();
-            if (error_state)
-              return octave_value_list ();
-            factorial (arg1, argInt, rnd);
-          }
+          factorial (arg1, rnd);
         else if (function == "gamma")
           evaluate (arg1, rnd, &mpfr_gamma);
         else if (function == "gammaln")
@@ -397,7 +443,7 @@ DEFUN_DLD (mpfr_function_d, args, nargout,
             return octave_value_list ();
           }
         break;
-        
+
       case 2: // binary function
         if      (function == "atan2")
           evaluate (arg1, arg2, rnd, &mpfr_atan2);
@@ -426,7 +472,7 @@ DEFUN_DLD (mpfr_function_d, args, nargout,
             return octave_value_list ();
           }
         break;
-        
+
       case 3: // ternary function
         if      (function == "fma")
           evaluate (arg1, arg2, arg3, rnd, &mpfr_fma);
@@ -446,4 +492,89 @@ DEFUN_DLD (mpfr_function_d, args, nargout,
 %!assert (mpfr_function_d ('plus', -inf, 1, eps / 2), 1);
 %!assert (mpfr_function_d ('plus', +inf, 1, eps / 2), 1 + eps);
 %!error mpfr_function_d ('Krauskefarben', 0, 47, 11);
+
+%!# Cross-check unit tests from crlibm against the MPFR library.
+%!# We simulate binary64 floating-point arithmetic in MPFR
+%!# with mpfr_function_d and results shall be identical.
+%!#
+%!shared testdata
+%! testdata = load (fullfile (...
+%!   fileparts (file_in_loadpath ("__check_crlibm__.m")), ...
+%!   "test", ...
+%!   "crlibm.mat"));
+
+%!function verify (fname, rnd, data)
+%!  assert (mpfr_function_d (fname, rnd, data.input), data.output);
+%!endfunction
+
+%!test verify ("acos", -inf, testdata.acos_rd);
+%!test verify ("acos", +inf, testdata.acos_ru);
+%!test verify ("acos",  0.5, testdata.acos_rn);
+%!test verify ("acos",  0,   testdata.acos_rz);
+
+%!test verify ("asin", -inf, testdata.asin_rd);
+%!test verify ("asin", +inf, testdata.asin_ru);
+%!test verify ("asin",  0.5, testdata.asin_rn);
+%!test verify ("asin",  0,   testdata.asin_rz);
+
+%!test verify ("atan", -inf, testdata.atan_rd);
+%!test verify ("atan", +inf, testdata.atan_ru);
+%!test verify ("atan",  0.5, testdata.atan_rn);
+%!test verify ("atan",  0,   testdata.atan_rz);
+
+%!test verify ("cos", -inf, testdata.cos_rd);
+%!test verify ("cos", +inf, testdata.cos_ru);
+%!test verify ("cos",  0.5, testdata.cos_rn);
+%!test verify ("cos",  0,   testdata.cos_rz);
+
+%!test verify ("cosh", -inf, testdata.cosh_rd);
+%!test verify ("cosh", +inf, testdata.cosh_ru);
+%!test verify ("cosh",  0.5, testdata.cosh_rn);
+%!test verify ("cosh",  0,   testdata.cosh_rz);
+
+%!test verify ("exp", -inf, testdata.exp_rd);
+%!test verify ("exp", +inf, testdata.exp_ru);
+%!test verify ("exp",  0.5, testdata.exp_rn);
+%!test verify ("exp",  0,   testdata.exp_rz);
+
+%!test verify ("expm1", -inf, testdata.expm1_rd);
+%!test verify ("expm1", +inf, testdata.expm1_ru);
+%!test verify ("expm1",  0.5, testdata.expm1_rn);
+%!test verify ("expm1",  0,   testdata.expm1_rz);
+
+%!test verify ("log", -inf, testdata.log_rd);
+%!test verify ("log", +inf, testdata.log_ru);
+%!test verify ("log",  0.5, testdata.log_rn);
+%!test verify ("log",  0,   testdata.log_rz);
+
+%!test verify ("log10", -inf, testdata.log10_rd);
+%!test verify ("log10", +inf, testdata.log10_ru);
+%!test verify ("log10",  0.5, testdata.log10_rn);
+%!test verify ("log10",  0,   testdata.log10_rz);
+
+%!test verify ("log1p", -inf, testdata.log1p_rd);
+%!test verify ("log1p", +inf, testdata.log1p_ru);
+%!test verify ("log1p",  0.5, testdata.log1p_rn);
+%!test verify ("log1p",  0,   testdata.log1p_rz);
+
+%!test verify ("log2", -inf, testdata.log2_rd);
+%!test verify ("log2", +inf, testdata.log2_ru);
+%!test verify ("log2",  0.5, testdata.log2_rn);
+%!test verify ("log2",  0,   testdata.log2_rz);
+
+%!test verify ("sin", -inf, testdata.sin_rd);
+%!test verify ("sin", +inf, testdata.sin_ru);
+%!test verify ("sin",  0.5, testdata.sin_rn);
+%!test verify ("sin",  0,   testdata.sin_rz);
+
+%!test verify ("sinh", -inf, testdata.sinh_rd);
+%!test verify ("sinh", +inf, testdata.sinh_ru);
+%!test verify ("sinh",  0.5, testdata.sinh_rn);
+%!test verify ("sinh",  0,   testdata.sinh_rz);
+
+%!test verify ("tan", -inf, testdata.tan_rd);
+%!test verify ("tan", +inf, testdata.tan_ru);
+%!test verify ("tan",  0.5, testdata.tan_rn);
+%!test verify ("tan",  0,   testdata.tan_rz);
+
 */
